/* ========================================================================
 * PlantUML : a free UML diagram generator
 * ========================================================================
 *
 * (C) Copyright 2009-2024, Arnaud Roques
 *
 * Project Info:  https://plantuml.com
 * 
 * If you like this project or if you find it useful, you can support us at:
 * 
 * https://plantuml.com/patreon (only 1$ per month!)
 * https://plantuml.com/paypal
 * 
 * This file is part of PlantUML.
 *
 * PlantUML is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * PlantUML distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library.  If not, see <https://www.gnu.org/licenses/>.
 *
 *
 * Original Author:  Arnaud Roques
 */
/*
 * Copyright (C) 2015 Square, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.plantuml.quantization;

import java.util.Comparator;
import java.util.WeakHashMap;

/**
 * Provides a stable ordering of objects, such that:
 *
 * <ul>
 * <li>compare(a, b) == 0 iff a == b</li>
 * <li>sign(compare(a, b)) == -sign(compare(b, a))</li>
 * </ul>
 *
 * <p>
 * Similar to Guava's {code Ordering.arbitrary()}.
 */
final class ArbitraryComparator implements Comparator<Object> {
	public static final ArbitraryComparator INSTANCE = new ArbitraryComparator();

	/**
	 * If we have no other way to order two objects in a stable manner, we will
	 * register both in this map and order them according to their associated
	 * values. The map's values are just integers corresponding to the order in
	 * which objects were added.
	 */
	private static final WeakHashMap<Object, Integer> objectIds = new WeakHashMap<>();

	private ArbitraryComparator() {
	}

	@Override
	public int compare(Object a, Object b) {
		if (a == b)
			return 0;
		if (a == null)
			return -1;
		if (b == null)
			return 1;

		int identityHashCodeDifference = System.identityHashCode(a) - System.identityHashCode(b);
		if (identityHashCodeDifference != 0) {
			return identityHashCodeDifference;
		}

		// We have an identityHashCode collision.
		return getObjectId(a) - getObjectId(b);
	}

	/**
	 * Get the ID of an object, adding it to the ID map if it isn't already
	 * registered.
	 */
	private static int getObjectId(Object object) {
		synchronized (objectIds) {
			Integer id = objectIds.get(object);
			if (id == null) {
				id = objectIds.size();
				objectIds.put(object, id);
			}
			return id;
		}
	}
}
